#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######       Module additionnel: arithmetique.py          ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.30]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
from __future__ import division

##################################################################
# Importation des modules système
#
import time
from sys import modules

###########################################################################
# Pour préparer la détection de la demande d'arrêt de calcul dans les calculs longs
#
prog=modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#
from math import *

###############################################################################
# Factorielle de n
# - version volontairement non récursive (la pile de python est limitée à environ 1000)
# - cette temporisation curieuse est justifiée. Si on la retire, la fonction ne s'arrête
# plus sur demande pour certaines valeurs de n (env. 20000). Et ce codage permet
#   de ne pas avoir la tempo à chaque boucle et donc diminue le temps de calcul
#
def fact(n):
    """fact(n): calcule la factorielle de n (entier >= 0)"""
    # Vérification de la donnée
    global prog
    # Vérification de la donnée
    if (not prog.estentier(n)) or (n<0):
        raise ValueError ("erreur fact(n): n doit être un nombre entier positif ou nul")
    x=1
    np1=n+1
    for i in xrange(2,np1):
        x*=i
        if i<=1000:
            pass
        elif i<=5000:
            if (i%50)==0:
                prog.testarretcalcul()
        elif i<=10000:
            if (i%10)==0:
                time.sleep(0.001)
                prog.testarretcalcul()
        else:
            time.sleep(0.002)
            prog.testarretcalcul()
    return x

###############################################################################
# Calcul du PGCD = Plus Grand Commun Diviseur entre 2 nombres entiers
#  par l'algorithme d'Euclide (version non récursive)
#
def pgcd(a,b):
    """pgcd(a,b): calcul du 'Plus Grand Commun Diviseur' entre les 2 nombres entiers a et b"""
    global prog
    # Verification des donnees
    if (not prog.estentier(a)) or (not prog.estentier(b)):
        raise ValueError ("erreur pgcd(a,b): a et b doivent être des nombres entiers")
    if ((a<=0) or (b<=0)):
        raise ValueError ("erreur pgcd(a,b): a et b doivent être > 0")
    # Calcul du pgcd
    x=a
    y=b
    while y<>0:
        # permettre l'arret du calcul s'il est demandé
        prog.testarretcalcul()
        r=x % y
        x=y
        y=r
    return x

###############################################################################
# Calcul du PPCM = Plus Petit Commun Multiple entre 2 nombres entiers
#  en utilisant le calcul du pgcd
#
def ppcm(a,b):
    """ppcm(a,b): calcul du 'Plus Petit Commun Multiple' entre 2 nombres entiers a et b"""
    global prog
    # Verification des donnees
    if (not prog.estentier(a)) or (not prog.estentier(b)):
        raise ValueError ("erreur ppcm(a,b): a et b doivent être des nombres entiers")
    if ((a<=0) or (b<=0)):
        raise ValueError ("erreur ppcm(a,b):  a et b doivent être > 0")
    # Calcul du ppcm
    return (a*b)//pgcd(a,b)

###############################################################################
# Décomposition d'un nombre entier n en facteurs premiers
# -> renvoie la liste des facteurs premiers trouvés
# -> si n est le seul facteur trouvé, il est un nombre premier
#
def facteurs(n):
    """facteurs(n): décomposition d'un nombre entier n en facteurs premiers"""
    global prog
    # Verification des donnees
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur facteurs(n): n doit être un nombre entier positif")
    # Calcul de la decomposition
    f=[]
    if n==1:
        return f
    i=2
    r=long(sqrt(n))+1
    while (i<=n):
        # permettre l'arret du calcul s'il est demandé
        prog.testarretcalcul()
        if i>r:
            f.append(n)
            break
        x=divmod(n,i)
        if x[1]==0:
            f.append(i)
            n=x[0]
            r=long(sqrt(n))+1
        else:
            i+=1
    return f

###############################################################################
# Dit si un nombre est premier
# -> renvoie "True" si c'est vrai, "False" sinon
# - par exemple: estpremier(999999937) renvoie "True"
# vérif: beaucoup de nb entiers ici: http://nombrespremiersliste.free.fr/
#
def estpremier(n):
    """estpremier(n): dit si un nombre est premier (renvoie True ou False)"""
    global prog
    # Verification des donnees
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur estpremier(n): n doit être un nombre entier positif")
    # Calcul du test
    if n<7:
        if n in (2,3,5):
            return True
        else:
            return False
    # si n est pair et >2 (=2: cas traité ci-dessus), il ne peut pas être premier
    if (n % 2)==0:
        return False
    k=7
    r=sqrt(n)+1
    while k<=r:
        # permettre l'arret du calcul s'il est demandé
        prog.testarretcalcul()
        if (n % k)==0:
            return False
        k+=2
    return True

###############################################################################
# Trouve tous les nombres premiers <= n à partir de 2
# => renvoie la liste. par exemple: premiers(10) => [2, 3, 5, 7]
# - chaque nouveau nb k testé n'est divisé que par les nb premiers déjà trouvés
# - au delà de 7 (qui est premier), on ne teste que les nb impairs
# - arrêt à racine de k: si on ne l'a pas trouvé avant, on ne le trouvera plus
# le temps de calcul augmente très rapidement avec n
# vérif: beaucoup de nb premiers ici: http://nombrespremiersliste.free.fr/
#
def premiers(n):
    """premiers(n): trouve tous les nombres premiers <= n et les affiche sous forme d'une liste"""
    global prog
    # Vérification des données
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur premiers(n): n doit être un nombre entier positif")
    # Calcul de la liste des nb premiers <= n
    if n==1:
        return []
    if n==2:
        return [2]
    if n in (3,4):
        return [2,3]
    if n in (5,6):
        return [2,3,5]
    p=[2,3,5]
    k=7
    while k<=n:
        # permettre l'arret du calcul s'il est demandé
        prog.testarretcalcul()
        # calcul
        rk=int(sqrt(k))+1
        i=0
        while i<len(p):
            if p[i]>rk:
                p.append(k)
                break
            if (k % p[i])==0:
                break
            i+=1
        k+=2
    return p

