#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######          Module additionnel: bibgene.py            ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.30]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules système
#
import time
from sys import modules

###############################################################################
# Pour préparer la détection de la demande d'arrêt de calcul dans les calculs longs
#
prog=modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#
from math import *
from copy import *
from decimal import *

###############################################################################
# dit si une chaîne binaire est valide
#  si L est donnée (>0), C doit être de longueur L
#  C doit être une chaîne non vide ne contenant que des "0" et des "1"
#
def okbin(C, L=0):
    """okbin(C, L=0): dit si une chaîne binaire est valide. Renvoie True ou False"""
    if type(C)!=str:
        return False
    lg=len(C)
    if lg==0:
        return False
    if not prog.estentier(L):
        return False
    if (L>0) and (lg!=L):
        return False
    if (C.count("0")+C.count("1"))!=lg:
        return False
    return True

###############################################################################
# Inversion des bits d'une chaîne binaire de longueur quelconque
#
def invbin(C):
    """invbin(C): inversion des bits d'une chaîne binaire de longueur quelconque"""
    if not okbin(C):
        raise ValueError ("erreur invbin(C): mauvaise chaîne binaire C")
    L=[]
    for i in range (0,len(C)):
        if C[i]=="0":
            L.append("1")
        else:
            L.append("0")
    ch=""
    return ch.join(L)

###############################################################################
# Complément à 2 d'une chaîne binaire de longueur quelconque
#
def compbin2(C):
    """compbin2(C): complément à 2 d'une chaîne binaire de longueur quelconque"""
    # Vérif des données
    if not okbin(C):
        raise ValueError ("erreur compbin2(C): mauvaise chaîne binaire C")
    # Calcul
    return dec2bin(bin2dec(invbin(C))+1)

###############################################################################
# Conversion nombre entier positif -> chaîne binaire (si B>0, complète à gauche par des zéros)
#
def dec2bin(D,B=0):
    """dec2bin(D,B=0): conversion nombre entier positif -> chaîne binaire (si B>0, complète à gauche par des zéros)"""
    global prog
    # Vérif des données
    if (not prog.estentier(D)) or (D<0):
        raise ValueError ("erreur dec2bin(D,B=0): D doit être un entier positif ou nul")
    if (not prog.estentier(B)) or (B<0):
        raise ValueError ("erreur dec2bin(D,B=0): B doit être un entier positif ou nul")
    if (B>0) and (D>=(1<<B)):
            raise ValueError ("erreur dec2bin(D,B=0): D trop grand pour le B donné")
    # Calcul
    dg="01"
    z=""
    while D>=2:
        x=divmod(D,2)
        z=dg[x[1]]+z
        D=x[0]
    z=dg[D]+z
    # retourne z, éventuellement complété par des zéros à gauche
    return z.zfill(B)

###############################################################################
# Conversion nombre entier signé -> chaîne binaire pour un mot de B bits (8 par défaut)
#  Un nombre négatif est représenté par le complément à 2 de sa valeur absolue
#
def dec2bins(D,B=8):
    """dec2bins(D,B): conversion nombre entier signé -> chaîne binaire pour un mot de B bits """
    global prog
    # Vérif des données
    if (not prog.estentier(D)):
        raise ValueError ("erreur dec2bins(D,B=8): D doit être un entier")
    if (not prog.estentier(B)) or (B<0):
        raise ValueError ("erreur dec2bins(D,B=8): B doit être un entier")
    if B<2:
        raise ValueError ("erreur dec2bins(D,B=8): B trop faible pour un binaire signé ou négatif")
    if (D>((1<<(B-1))-1)) or (D<-(1<<(B-1))):
        raise ValueError ("erreur dec2bins(D,B=8): nombre D trop fort pour un mot de B bits signé")
    # Calcul
    if D<0:
        return dec2bin((1<<B)+D,B)
    else:
        return dec2bin(D,B)

###############################################################################
# Conversion chaîne binaire de longueur quelconque -> nombre entier positif
#
def bin2dec(C):
    """bin2dec(C): Conversion chaîne binaire de longueur quelconque -> nombre entier positif"""
    # vérif des données
    if not okbin(C):
        raise ValueError ("erreur bin2dec(C): mauvaise chaîne binaire C")
    # Calcul
    return int(C,2)

###############################################################################
# Conversion chaîne binaire de longueur B bits -> nombre entier signé
#
def bin2decs(C,B=8):
    """bin2decs(C,B): Conversion chaîne binaire de longueur B bits -> nombre entier signé"""
    # vérif des données
    if not okbin(C):
        raise ValueError ("erreur bin2decs(C,B): mauvaise chaîne binaire C")
    if B<2:
        raise ValueError ("erreur bin2decs(C,B): mauvais paramètre B<2 pour un binaire signé")
    # Calcul
    # retire les éventuels zéro en trop à gauche
    while (len(C)>B) and (C[0]=="0"):
        C=C[1:]
    # complète éventuellement le mot de B bits par des zéros à gauche
    C=C.zfill(B)
    if not okbin(C,B):
        raise ValueError ("erreur bin2decs(C,B): mauvaise chaîne binaire C de B bits")
    # Calcul de la valeur positive
    D=int(C,2)
    # calcul de la valeur si négative
    if C[0]=="1":
        # le 1er chiffre est un '1' => le résultat sera négatif
        return D-(1<<B)
    else:
        # le 1er chiffre est un '0' => le résultat sera positif
        return D

###############################################################################
# Conversion degrés décimaux -> degrés sous la forme d'une liste: [degrés, minutes, secondes]
#
def deg2dms(DD):
    """deg2dms(DD): conversion degrés décimaux -> degrés sous la forme d'une liste: [degrés, minutes, secondes]"""
    global prog
    # vérif des données
    if (not prog.estnombre(DD)):
        raise ValueError ("erreur deg2dms(DD): DD doit être un nombre")
    # Calcul
    L=[]
    x=int(DD)
    L.append(x)
    m=(DD-x)*60
    x=int(m)
    L.append(x)
    s=(m-x)*60
    L.append(s)
    return L

###############################################################################
# Conversion degrés sous la forme d'une liste [degrés, minutes, secondes] -> en degrés décimaux
#
def dms2deg(L):
    """dms2deg(L): conversion degrés sous la forme d'une liste [degrés, minutes, secondes] -> en degrés décimaux"""
    global prog
    # vérif des données
    if (type(L)!=list) or (len(L)!=3) or (not prog.estnombre(L[0])) or (not prog.estnombre(L[1])) or (not prog.estnombre(L[2])):
        raise ValueError ("erreur dms2deg(L): L doit être une liste de 3 nombres")
    # Calcul
    return L[0] + (L[1]/60) + (L[2]/3600)

###############################################################################
# Construit une liste de n paramètres ou de n listes de paramètres de L1 (début) à L2 (fin)
#  si L1 et L2 sont des nombres, donne le résultat sous forme de liste de n valeurs
#  si L1 et L2 sont des listes, donne le résultat sous forme de liste de n listes de valeurs
#
def jeuxdessais(n,L1,L2):
    """jeudessais(n,L1,L2): Construit une liste de n listes de paramètres de L1 (début) à L2 (fin)"""
    global prog
    # vérification de la validité de n
    if (not prog.estentier(n)) or (n<2):
        raise ValueError ("erreur jeudessais(n,L1,L2): n doit être un nombre entier >= 2")
    L=[]
    if type(L1)!=list:
        # si L1 n'est pas une liste, c'est forcément un nombre
        if (not prog.estnombre(L1)):
            raise ValueError ("erreur jeudessais(n,L1,L2): L1 doit être soit un nombre soit une liste")
        # ici, L1 est un nombre, alors L2 doit en être un aussi
        if (not prog.estnombre(L2)):
            raise ValueError ("erreur jeuxdessais(n,L1,L2), comme L1 est un nombre, L2 doit en être un aussi")
        # calcul d'une liste simple (et non d'une liste de liste)
        L2m1=L2-L1
        for i in range(0,n):
            L.append(L1+L2m1/(n-1)*i)
    else:
        # L1 est une liste, L2 doit en être une aussi!
        if type(L2)!=list:
            raise ValueError ("erreur jeuxdessais(n,L1,L2), comme L1 est une liste, L2 doit en être une aussi")
        if L1==[]:
            raise ValueError ("erreur jeuxdessais(n,L1,L2), L1 doit être une liste non vide")
        np=len(L1)
        if np!=len(L2):
            raise ValueError ("erreur jeuxdessais(n,L1,L2), L1 et L2 n'ont pas le même nombre de membres")
        # calcul d'une liste de n listes de np paramètres compris entre L1 et L2
        for i in range(0,n):
            Lp=[]
            for k in range(0,np):
                Lp.append(L1[k]+(L2[k]-L1[k])/(n-1)*i)
            L.append(Lp)
    return L

###############################################################################
# Calcule le volume restant de liquide d'une cuve cylindique horizontale
#  en fonction de son diamètre d, de son volume total vt et du niveau du liquide h
#  exemple d'application: cuve de mazout
#
def volcuve(h,d,vt):
    """volcuve(h,d,vt): Calcule le volume restant de liquide d'une cuve cylindique horizontale de diamètre d et de volume total vt"""
    global prog
    if (not prog.estnombre(h)) or (h<0):
        raise ValueError ("erreur volcuve(h,d,vt): h doit être un nombre positif ou nul")
    if (not prog.estnombre(d)) or (d<=0):
        raise ValueError ("erreur volcuve(h,d,vt): d doit être un nombre positif")
    if (not prog.estnombre(vt)) or (vt<=0):
        raise ValueError ("erreur volcuve(h,d,vt): vt doit être un nombre positif")
    if h>d:
        raise ValueError ("erreur volcuve(h,d,vt): h ne peut pas être supérieur à d (=ça déborde!)")
    r=d/2
    if h==d:
        v=vt
    elif h==r:
        v=vt/2
    else:
        st=pi*r*r
        if h<r:
            alpha=acos((r-h)/r)
            sr=(r-h)*sqrt(r*r-(r-h)**2)
            sa=st/pi*alpha
            sh=sa-sr
        else:
            alpha=acos((h-r)/r)
            sr=(h-r)*sqrt(r*r-(h-r)**2)
            sa=st/pi*(pi-alpha)
            sh=sa+sr
        v=sh/st*vt
    return v

###############################################################################
# Convertit en chaîne hexadécimale les mots de passe donnés en texte clair
#  application: pour les mots de passe utilisés en wifi
#
def mdp2hex(ch):
    """mdp2hex(ch): Convertit en chaîne hexadécimale les mots de passe donnés en texte clair"""
    if type(ch)!=str:
        raise ValueError ("erreur mdp2hex(ch): ch doit être une chaîne de caractères")
    if ch=="":
        return ""
    h=""
    for i in range(0,len(ch)):
        h+=str(hex(ord(ch[i])))[2:]
    return h

###############################################################################
# Convertit en texte clair les mots de passe donnés en chaîne hexadécimale
#  application: pour les mots de passe utilisés en wifi
#
def hex2mdp(ch):
    """hex2mdp(ch): Convertit en texte clair les mots de passe donnés en chaîne hexadécimale"""
    if type(ch)!=str:
        raise ValueError ("erreur hex2mdp(ch): ch doit être une chaîne de caractères")
    if ch=="":
        return ""
    if (len(ch)%2)!=0:
        raise ValueError ("erreur hex2mdp(ch): ch doit avoir un nombre pair de caractères")
    for i in range(0,len(ch)):
        if ch[i] not in "0123456789ABCDEFabcdef":
            raise ValueError ("erreur hex2mdp(ch): la chaîne hexadécimale ch a des caractères anormaux")
    h=""
    for i in range(0,len(ch),2):
        h+=chr(int(ch[i:i+2],16))
    return h

