#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######       Module additionnel: combinatoire.py          ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.30]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules système
#
from sys import modules
import time

###############################################################################
# Pour préparer la détection de la demande d'arrêt de calcul dans les calculs longs
#
prog=modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#
from math import *

###############################################################################
# Nombre de permutations de n objets en tenant compte de l'ordre
# NB: par convention: permut(0)=1 (mais pas de sens physique)
# - version volontairement non récursive (la pile de python est limitée à environ 1000)
# - cette temporisation curieuse est justifiée. Si on la retire, la fonction ne s'arrête
# plus sur demande pour certaines valeurs de n (env. 20000). Et ce codage permet
#   de ne pas avoir la tempo à chaque boucle et donc diminue le temps de calcul
#
def permut(n):
    """permut(n): nombre de permutations de n objets en tenant compte de l'ordre"""
    global prog
    # Vérification de la donnée
    if (not prog.estentier(n)) or (n<0):
        raise ValueError ("erreur permut(n): n doit être un nombre entier positif ou nul")
    # Calcul
    x=1
    np1=n+1
    for i in xrange(2,np1):
        x*=i
        if i<=1000:
            pass
        elif i<=5000:
            if (i%50)==0:
                prog.testarretcalcul()
        elif i<=10000:
            if (i%10)==0:
                time.sleep(0.001)
                prog.testarretcalcul()
        else:
            time.sleep(0.002)
            prog.testarretcalcul()
    return x

###############################################################################
# Arrangement de n objets pris p a p
#
def arrang(n,p):
    """arrang(n,p): arrangement de n objets pris p a p"""
    global prog
    # Vérification des données
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur arrang(n,p): n doit être un nombre entier positif")
    if (not prog.estentier(p)) or (p<0):
        raise ValueError ("erreur arrang(n,p): p doit être un nombre entier positif ou nul")
    if (p>n):
        raise ValueError ("erreur arrang(n,p): p doit être inférieur ou égal à n")
    # Calcul
    x=1
    i=n-p+1
    while i<=n:
        # permettre l'arret du calcul s'il est demandé
        prog.testarretcalcul()
        x*=i
        i+=1
    return x

###############################################################################
# Combinaison de n objets pris p a p
# NB: par convention, comb(n,0)=1 (mais pas de sens physique)
#
def combin(n,p):
    """combin(n,p): combinaison de n objets pris p a p"""
    global prog
    # Vérification des données
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur combin(n,p): n doit être un nombre entier positif")
    if (not prog.estentier(p)) or (p<0):
        raise ValueError ("erreur combin(n,p): n doit être un nombre entier positif ou nul")
    if (p>n):
        raise ValueError ("erreur combin(n,p): p doit être inférieur ou égal à n")
    # Calcul
    return arrang(n,p)//permut(p)

