#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######       Module additionnel: probabilite.py           ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.30]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules système
#
from sys import modules
import time

###############################################################################
# Pour préparer la détection de la demande d'arrêt de calcul dans les calculs longs
#
prog=modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#   (à compléter si nécessaire)
#
from math import *
from random import *
from combinatoire import *


###############################################################################
# Loi binomiale
#   probabilité d'avoir k réussite(s) dans n évènements indépendants,
#     chaque évènement ayant une probabilité p de réussite.
#
def binom(k,n,p):
    """binom(k,n,p): probabilité d'avoir k réussite(s) dans n évènements indépendants, chaque évènement ayant une probabilité p% de réussite"""
    global prog
    if (not prog.estentier(k)) or (k<0):
        raise ValueError ("erreur binom(k,n,p): k doit être un entier positif ou nul")
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur binom(k,n,p): n doit être un entier positif")
    if (not prog.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur binom(k,n,p): p doit être un nombre compris entre 0 et +1")
    if (k>n):
        raise ValueError ("erreur binom(k,n,p): k doit être <= n")
    x=combin(n,k)*pow(p,k)*pow(1-p,n-k)
    return x

###############################################################################
# Loi binomiale cumulée
#   prob d'avoir 0,1,..,k réussite(s) dans n évènements indépendants,
#     chaque évènement ayant une probabilité p de réussite.
#
def binomcum(k,n,p):
    """binomcum(k,n,p): probabilité d'avoir 0,1,..,k réussite(s) dans n évènements indépendants, chaque évènement ayant une probabilité p% de réussite"""
    x=0
    for i in xrange(0,k+1):
        x+=binom(i,n,p)
    return x

###############################################################################
# Loi hypergéométrique
#   probabilité d'avoir k réussite(s) dans n évènements indépendants,
#     sachant qu'il y a g de réussites dans la population de taille t.
#
def hypgeo(k,n,g,t):
    """hypgeo(k,n,g,t): prob d'avoir k réussites dans un échantillon de taille n, sachant qu'il y en a g la population de taille t"""
    global prog
    if (not prog.estentier(k)) or (k<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): k doit être un entier positif ou nul")
    if (not prog.estentier(n)) or (n<=0):
        raise ValueError ("erreur hypgeo(k,n,p,t): n doit être un entier positif")
    if (not prog.estentier(g)) or (g<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): g doit être un entier positif ou nul")
    if (not prog.estentier(t)) or (t<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): t doit être un entier positif")
    if (k>n):
        raise ValueError ("erreur hypgeo(k,n,g,t): k doit être <= n")
    if (n>t):
        raise ValueError ("erreur hypgeo(k,n,g,t): n doit être <= t")
    if (g>t):
        raise ValueError ("erreur hypgeo(k,n,g,t): g doit être <= t")
    if (k>g):
        raise ValueError ("erreur hypgeo(k,n,g,t): k doit être <= g")
    return combin(g,k)*combin(t-g,n-k)/combin(t,n)

###############################################################################
# Loi hypergéometrique cumulée
#   probabilité d'avoir 0,1,2,...,k réussite(s) dans n évènements indépendants,
#     sachant qu'il y a g de réussites dans la population de taille t.
#
def hypgeocum(k,n,g,t):
    """hypgeocum(k,n,p,t): prob d'avoir 0,1,..,k réussites dans un échantillon de taille n, sachant qu'il y p% de réussite dans la population de taille t"""
    x=0
    for i in range(0,k+1):
        x+=hypgeo(i,n,g,t)
    return x

###############################################################################
# probabilité qu'une variable aléatoire distribuée selon une loi normale réduite soit inférieure à x
# au delà de x=7.5, renvoie 1.0
#
def pgaussred(x):
    """pgaussred(x): probabilité qu'une variable aléatoire distribuée selon une loi normale réduite soit inférieure à x"""
    global prog
    if not prog.estnombre(x):
        raise ValueError ("erreur pgaussred(x): x doit être un nombre")
    if x==0:
        return 0.5
    if x>8:
        return 1.0
    if x<-8:
        return 0.0
    u=abs(x)
    n=int(u*2000)
    du=u/n
    k=1/sqrt(2*pi)
    u1=0
    f1=k
    p=0.5
    for i in range(0,n):
        u2=u1+du
        f2=k*exp(-0.5*u2*u2)
        p=p+(f1+f2)*du*0.5
        u1=u2
        f1=f2
    if p>1.0:
        p=1.0
    if x>0:
        return p
    else:
        return 1.0-p

###############################################################################
# Loi normale (m, s)
#   prob qu'une variable aléatoire distribuée selon ln(m,s) soit inférieure à x
#
def pgauss(x,m,s):
    """pgauss(x): probabilité qu'une variable aléatoire distribuée selon une loi normale(m,s) soit inférieure à x"""
    global prog
    if not prog.estnombre(x):
        raise ValueError ("erreur pgauss(x,m,s): p doit être un nombre")
    if not prog.estnombre(m):
        raise ValueError ("erreur pgauss(x,m,s): m doit être un nombre")
    if (not prog.estnombre(s)) or (s<=0):
        raise ValueError ("erreur pgauss(x,m,s): s doit être un nombre positif non nul")
    return pgaussred((x-m)/s)

###############################################################################
# renvoie la variable x distribuée selon une LNR, et correspondant à une probabilité donnée p
#
def xgaussred(p):
    """xgaussred(p): renvoie la variable x distribuée selon une LNR, et correspondant à une probabilité donnée p"""
    global prog
    # vérification des données
    if (not prog.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur xgaussred(p): p doit être un nombre >=0 et <=1")
    # calcul
    if p==0.5:
        return 0.0
    if p==0.0:
        return -8
    if p==1.0:
        return +8
    if p>0.5:
        pc=p
    else:
        pc=1-p
    du=1/2000
    eps=1.0E-15
    k=1/sqrt(2*pi)
    u1=0.0
    f1=k
    p1=0.5
    while True:
        # permettre l'arret du calcul s'il est demandé
        prog.calc.arretcalcul()
        u2=u1+du
        f2=k*exp(-0.5*u2*u2)
        s=(f1+f2)*0.5*du
        p2=p1+s
        if abs(p2-pc)<eps:
            break
        if ((p1<pc) and (p2<p1)) or  ((p1>pc) and (p2>p1)):
            # =on est en train de s'éloigner du point recherché
            du=-du/2
        u1=u2
        f1=f2
        p1=p2
    if p>0.5:
        return u2
    else:
        return -u2

###############################################################################
# renvoie la variable x distribuée selon une LN(m,s), et correspondant à une probabilité donnée p
#
def xgauss(p,m,s):
    """xgauss(p,m,s): renvoie la variable x distribuée selon une LN(m,s), et correspondant à une probabilité donnée p"""
    global prog
    if (not prog.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur xgauss(p,m,s): p doit être un nombre >=0 et <=1")
    if not prog.estnombre(m):
        raise ValueError ("erreur pgauss(x,m,s): m doit être un nombre")
    if not prog.estnombre(s):
        raise ValueError ("erreur pgauss(x,m,s): s doit être un nombre")
    return m+s*xgaussred(p)

###############################################################################
# renvoie une valeur au hasard de la variable x distribuée selon une LN réduite
#
def hgaussred(n=1):
    """hgaussred(): renvoie une valeur au hasard de la variable x distribuée selon une LN réduite"""
    global prog
    if (not prog.estentier(n)) or (n<1):
        raise ValueError ("erreur hgauss(m,s,n=1): s doit être un nombre entier >=1")
    if n==1:
        return xgaussred(random())
    else:
        L=[]
        for i in range(0,n):
            L.append(xgaussred(random()))
        return L

###############################################################################
# renvoie une valeur au hasard de la variable x distribuée selon une LN(m,s)
#
def hgauss(m,s,n=1):
    """hgauss(m,s): renvoie une valeur au hasard de la variable x distribuée selon une LN(m,s)"""
    global prog
    if not prog.estnombre(m):
        raise ValueError ("erreur hgauss(m,s,n=1): m doit être un nombre")
    if (not prog.estnombre(s)) or (s<=0):
        raise ValueError ("erreur hgauss(m,s,n=1): s doit être un nombre positif non nul")
    if (not prog.estentier(n)) or (n<1):
        raise ValueError ("erreur hgauss(m,s,n=1): s doit être un nombre entier >=1")
    if n==1:
        return xgauss(random(),m,s)
    else:
        L=[]
        for i in range(0,n):
            L.append(xgauss(random(),m,s))
        return L

