#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######          Module additionnel: temps.py              ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.30]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules système
#
from sys import modules
import locale
import time
import datetime
from socket import *
import struct

###############################################################################
# Pour préparer la détection de la demande d'arrêt de calcul dans les calculs longs
#
prog=modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#   (à compléter si nécessaire)
#
from math import *

###############################################################################
# Commande dont l'exécution est nécessaire lors du chargement du module
# pour avoir la présentation des dates en francais
#
locale.setlocale(locale.LC_TIME,'')

###############################################################################
# Fonctions dates et heures
###############################################################################

###############################################################################
# Donne la date et l'heure du jour ainsi que le jour de semaine
# sous le format ex: 'vendredi 19/10/2007 05:56:35'
#
def tempsordi():
    """tempsordi(): Donne la date, l'heure et le jour de semaine de l'ordinateur"""
    return  time.strftime('%A %d/%m/%Y %H:%M:%S')

###############################################################################
# Donne la date et l'heure exacte (à la seconde) par consultation d'un serveur ntp
#   nécessite donc un accès internet
# serveur ntp par defaut: ntp.univ-lyon1.fr
# autres serveurs France: voir http://www.cru.fr/services/ntp/serveurs_francais
# exemple de réponse de la fonction: 'samedi 10/11/2007 9:12:51'
#   (vous pourriez avoir aussi le numéro du jour dans l'année avec "str(t[7]))"
# vérif: l'heure est bien exacte à la sec près par rapport à une horloge "radio"
#
def tempsntp(sntp='ntp.univ-lyon1.fr'):
    """tempsntp(sntp='ntp.univ-lyon1.fr'): Donne la date et l'heure exacte par consultation d'un serveur ntp"""
    jsem=["lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi", "dimanche"]
    # nb de sec entre 1/1/1900 (=temps ntp) et 1/1/1970 (=temps python (=epoch))
    #    (calcul ici par: diffdate("1/1/1900","1/1/1970")*24*3600)
    # Vérification des données
    if type(sntp)!=str:
        raise ValueError ("erreur tempsntp(sntp='ntp.univ-lyon1.fr'): sntp doit être une chaine")
    # Calcul
    temps19701900 = 2208988800L
    buffer=1024
    # initialisation d'une connexion UDP
    client=socket(AF_INET, SOCK_DGRAM)
    # envoie de la requête UDP
    data='\x1b' + 47 * '\0'
    client.sendto(data, (sntp, 123))
    # réception de la réponse UDP
    data, addresse = client.recvfrom(buffer)
    if data:
        tps = struct.unpack('!12I', data)[10]
        tps -= temps19701900
        t=time.localtime(tps)
        ch=jsem[t[6]]+' '+str(t[2])+'/'+str(t[1])+'/'+str(t[0])+' '+str(t[3])+':'+str(t[4])+':'+str(t[5])
        return ch
    else:
        return "échec: pas de réponse du serveur"

###############################################################################
# Fonctions heures
###############################################################################

###############################################################################
# Donne l'heure courante sous le format ex: '05:57:56'
#
def cetteheure():
    """cetteheure(): donne l'heure courante de l'ordinateur sous le format ex: '05:57:56'"""
    return time.strftime('%X')

###############################################################################
# Donne l'heure en "hh:mm:ss.sss" a partir des secondes decimales
#
def sec2hms(sd):
    """sec2hms(sd): donne l'heure en "hh:mm:ss.sss" à partir des secondes décimales"""
    global prog
    # Vérification de la donnée
    if (not prog.estnombre(sd)) or (sd<0):
        raise ValueError ("erreur sec2hms(sd): sd doit être un nombre positif ou nul")
    # Calcul
    h=0
    m=0
    s=sd
    if s>60:
        m=int(s/60)
        s=s-m*60
        if m>60:
            h=int(m/60)
            m=m-h*60
    hch=str(h)
    hch=hch.zfill(2)
    mch=str(m)
    mch=mch.zfill(2)
    sch=str(round(s,3))
    if s<10:
        sch="0"+sch
    return hch+":"+mch+":"+sch

###############################################################################
# Fonctions dates
###############################################################################

###############################################################################
# Donne la date du jour sous le format 'jj/mm/aaaa' (ex: '20/10/2007')
#
def cejour():
    """cejour(): donne la date du jour sous le format 'jj/mm/aaaa' (ex: '20/10/2007')"""
    return time.strftime('%d/%m/%Y')

###############################################################################
# Dit si l'année donnée "a" (aaaa) est bissextile ou non
#
def bissextile(a):
    """bissextile(a): dit si l'année donnée est bissextile ou non (True=oui, False=non)"""
    global prog
    # Vérification de la donnée
    if (not prog.estentier(a)) or (a<=0):
        raise ValueError ("erreur bissextile(a): a doit être un nombre entier positif")
    # Calcul
    if (a % 4)==0:
        if ((a % 100)==0) and ((a % 400)<>0):
            return False
        else:
            return True
    else:
        return False

###############################################################################
# Vérifie si une date liste [j,m,a] est valide, et renvoie True ou False
#
def okdate(lst):
    """okdate(lst): vérifie si une date liste [j,m,a] est valide, et renvoie True si oui ou False sinon"""
    nbjmn=[31,28,31,30,31,30,31,31,30,31,30,31]
    nbjmb=[31,29,31,30,31,30,31,31,30,31,30,31]
    if type(lst)<>list:
        # le paramètre lst n'est pas une liste
        return False
    if len(lst)<>3:
        # la liste n'a pas les 3 nombres attendus
        return False
    j=lst[0]
    m=lst[1]
    a=lst[2]
    if (int(a)<>a) or (int(m)<>m) or (int(j)<>j):
        # l'un des paramètres n'est pas un nombre entier
        return False
    if (m<1) or (m>12):
        # le mois n'est pas bon
        return False
    if bissextile(a):
        nbjm=nbjmb
    else:
        nbjm=nbjmn
    k=nbjm[m-1]
    if (j<1) or (j>k):
        # le jour n'est pas bon
        return False
    if a<1:
        # l'année n'est pas bonne
        return False
    # la date est ok (ouf...)
    return True

###############################################################################
# Transforme une chaîne date "j/m/a" en une liste [j,m,a] et la vérifie
#
def saisiedate(D):
    """saisiedate(D): transforme la date D ('j/m/a') en une liste [j,m,a] et la vérifie """
    d=("0","1","2","3","4","5","6","7","8","9")
    lst=[]
    # verif chaîne
    if type(D)<>str:
        raise ValueError ("erreur saisiedate(D): D doit être une chaîne")
    for i in range(0,len(D)):
        if (D[i] not in d) and (D[i]<>"/"):
            raise ValueError ("erreur saisiedate(D): caractères anormaux dans D")
    k=D.count("/")
    if k<>2:
        raise ValueError ("erreur saisiedate(D): il faut 2 séparateurs '/'")
    # recherche du jour
    i=D.find("/")
    if i==0:
        raise ValueError ("erreur saisiedate(D): manque le jour")
    Dx=D[0:i]
    while Dx[0]=="0":
        Dx=Dx[1:len(D)]
    lst.append(int(Dx))
    # recherche du mois
    j=D.find("/",i+1)
    if j==i+1:
        raise ValueError ("erreur saisiedate(D): manque le mois")
    Dx=D[i+1:j]
    while Dx[0]=="0":
        Dx=Dx[1:len(D)]
    lst.append(int(Dx))
    # recherche de l'année
    Dx=D[j+1:len(D)]
    if len(Dx)==0:
        raise ValueError ("erreur saisiedate(D): manque l'année")
    while Dx[0]=="0":
        Dx=Dx[1:len(D)]
    lst.append(int(Dx))
    # vérif validité de la date
    if okdate(lst)==False:
        raise ValueError ("erreur saisiedate(D): date D invalide")

    return lst

###############################################################################
# Transforme une date liste [j,m,a] en une chaîne 'jj/mm/aaaa' pour affichage
#  suppose que la date est valide et ne la vérifie pas
#
def affichedate(lst):
    """affichedate(lst): transforme une date liste=[j,m,a] en une chaîne 'jj/mm/aaaa'"""
    j=lst[0]
    if j<10:
        D="0" + str(j)
    else:
        D=str(j)
    D=D + "/"
    m=lst[1]
    if m<10:
        D=D + "0" + str(m)
    else:
        D=D + str(m)
    D=D + "/"
    a=lst[2]
    Dx=str(a)
    while len(Dx)<4:
        Dx="0"+Dx
    D=D+Dx
    return D

###############################################################################
# Normalise une date chaîne 'j/m/a' pour la mettre sous la forme 'jj/mm/aaaa' et la vérifie
#
def normaldate(D):
    """normaldate(D): normalise et vérifie une date 'j/m/a' => 'jj/mm/aaaa'"""
    return affichedate(saisiedate(D))

###############################################################################
# Donne le nombre de jour de l'annee a
#
def nbjoursan(a):
    """nbjoursan(a): donne le nombre de jours de l'année"""
    global prog
    # Vérification de la donnée
    if (not prog.estentier(a)) or (a<=0):
        raise ValueError ("erreur nbjoursan(a): a doit être un nombre entier positif")
    # Calcul
    if bissextile(int(a)):
        return 366
    else:
        return 365

###############################################################################
# Donne le numéro du jour dans l'année d'une date D 'j/m/a' (1er janvier = 1, ...)
#
def numjouran(D):
    """numjouran(D): donne le numéro du jour dans l'année  de la date D 'j/m/a' (1er janvier = 1, ...)"""
    jmn=[0,31,59,90,120,151,181,212,243,273,304,334,365]
    jmb=[0,31,60,91,121,152,182,213,244,274,305,335,366]
    # Vérification des données
    if type(D)<>str:
        raise ValueError ("erreur numjouran(D): D doit être une chaîne")
    # Calcul
    lst=saisiedate(D)
    if bissextile(lst[2]):
        jm=jmb
    else:
        jm=jmn
    return jm[lst[1]-1]+lst[0]

###############################################################################
# Donne la date qui est le nième jour de l'année a
#
def datenumjouran(n,a):
    """datenumjouran(n,a): donne la date qui est le nième jour de l'année a"""
    global prog
    jmn=[0,31,59,90,120,151,181,212,243,273,304,334,365]
    jmb=[0,31,60,91,121,152,182,213,244,274,305,335,366]
    # Vérification des données
    if (not prog.estentier(n)) or (n<0):
        raise ValueError ("erreur datenumjouran(n,a): n doit être un nombre entier positif ou nul")
    if (not prog.estentier(a)) or (a<=0):
        raise ValueError ("erreur datenumjouran(n,a): a doit être un nombre entier positif")
    # Calcul
    if bissextile(a):
        jm=jmb
    else:
        jm=jmn
    for i in range(1,13):
        if jm[i]>=n:
            m=i
            break
    j=n-jm[m-1]
    return affichedate([j,m,a])

###############################################################################
# Dit si une date D2 'j/m/a' est postérieure ou égale à une autre date D1 'j/m/a'
#
def postdate(D1,D2):
    """postdate(D1,D2): dit si une date D2 'j/m/a' est postérieure ou égale à une autre date D1 'j/m/a'"""
    # Vérification des données
    if type(D1)<>str:
        raise ValueError ("erreur postdate(D1,D2): D1 doit être une chaîne")
    if type(D2)<>str:
        raise ValueError ("erreur postdate(D1,D2): D2 doit être une chaîne")
    # Calcul
    lst1=saisiedate(D1)
    lst2=saisiedate(D2)
    if (lst2[2]>lst1[2]):
        return True
    elif (lst2[2]==lst1[2]):
        if (lst2[1]>lst1[1]):
            return True
        elif (lst2[1]==lst1[1]):
            if (lst2[0]>=lst1[0]):
                return True
    return False

###############################################################################
# Donne le nombre de jours entre 2 dates 'j/m/a'
# Si D2 postérieure ou égale a D1, renvoie un nb positif ou nul
# sinon, renvoie un nb négatif
#
def diffdate(D1,D2):
    """diffdate(D1,D2): nombre de jours entre 2 dates ('j/m/a'), >=0 si D2>=D1, <0 sinon"""
    # Vérification des données
    if type(D1)<>str:
        raise ValueError ("erreur diffdate(D1,D2): D1 doit être une chaîne")
    if type(D2)<>str:
        raise ValueError ("erreur diffdate(D1,D2): D2 doit être une chaîne")
    # Calcul
    a1=saisiedate(D1)[2]
    a2=saisiedate(D2)[2]
    x1=numjouran(D1)
    x2=numjouran(D2)
    if postdate(D1,D2):
        # D2 est postérieure a D1: retour d'un nb >=0
        nk=x2-x1
        for ak in range(a1,a2):
            nk=nk+nbjoursan(ak)
        return nk
    else:
        # D2 est postérieure a D1: retour d'un nb >=0
        nk=x2-x1
        for ak in range(a2,a1):
            nk=nk-nbjoursan(ak)
        return nk

###############################################################################
# Donne la date qui est postérieure de n jours à la date D 'j/m/a'
#   ou antérieure de n jours si n<0
#
def adddate(D,n):
    """adddate(D,n): donne la date qui est postérieure de n jours (n>=0) ou antérieure de n jours (n<0) à la date D 'j/m/a'"""
    global prog
    # Vérification des données
    if type(D)<>str:
        raise ValueError ("erreur adddate(D,n): D doit être une chaîne")
    if not prog.estentier(n):
        raise ValueError ("erreur adddate(D,n): n doit être un nombre entier")
    # Calcul
    ak=saisiedate(D)[2]
    if n>=0:
        nk=numjouran(D)+n
        while nk>nbjoursan(ak):
            nk=nk-nbjoursan(ak)
            ak=ak+1
        return datenumjouran(nk,ak)
    else:
        nk=abs(n)+nbjoursan(ak)-numjouran(D)
        while nk>=nbjoursan(ak):
            nk=nk-nbjoursan(ak)
            ak=ak-1
        return datenumjouran(nbjoursan(ak)-nk,ak)

###############################################################################
# Donne le numéro du jour de la semaine d'une date D 'j/m/a' (lundi=1, ...)
#
def numjoursem(D):
    """numjoursem(D): donne le numéro du jour de la semaine d'une date D 'j/m/a' (lundi=1, ...)"""
    # Vérification des données
    if type(D)<>str:
        raise ValueError ("erreur numjoursem(D): D doit être une chaîne")
    # Calcul
    a1=saisiedate(D)[2]-1
    j=(a1+(a1//4)-(a1//100)+(a1//400)+numjouran(D)) % 7
    if j==0: j=7
    return j

###############################################################################
# Donne le jour de la semaine d'une date D 'j/m/a' ('lundi' pour lundi, ...)
#
def joursem(D):
    """joursem(D): donne le jour de la semaine d'une date D 'j/m/a' ('lundi' pour lundi, ...)"""
    jsem=["lundi", "mardi", "mercredi", "jeudi", "vendredi", "samedi", "dimanche"]
    # Vérification des données
    if type(D)<>str:
        raise ValueError ("erreur joursem(D): D doit être une chaîne")
    # Calcul
    return jsem[numjoursem(D)-1]

###############################################################################
# Trouve la date du lundi d'une semaine qui contient une date D 'j/m/a'
#
def lundi(D):
    """lundi(D): trouve la date du lundi d'une semaine qui contient la date D 'j/m/a'"""
    # Vérification des données
    if type(D)<>str:
        raise ValueError ("erreur lundi(D): D doit être une chaîne")
    # Calcul
    x=numjoursem(D)
    return adddate(D,-x+1)

###############################################################################
# Trouve la date de Pâques d'une année donnée an
#  méthode Hodges: validité = calendrier grégorien de 1593 a 4100
#     (verif faite avec toutes les dates officielles)
#  voir méthode ici: http://users.chariot.net.au/~gmarts/eastalg.htm#EasterHodges
#  et pour vérif ici: http://www.dioceserimouski.com/ch/paquesdates.html
#
def paques(an):
    """paques(an): trouve la date de Pâques d'une année donnée an"""
    global prog
    # Vérification des données
    if (not prog.estentier(an)) or (an<=0):
        raise ValueError ("erreur paques(an): an doit être un nombre entier positif")
    # Calcul
    a=an//100
    b=an%100
    c=(3*(a+25))//4
    d=(3*(a+25))%4
    e=(8*(a+11))//25
    f=(5*a+b)%19
    g=(19*f+c-e)%30
    h=(f+11*g)//319
    j=(60*(5-d)+b)//4
    k=(60*(5-d)+b)%4
    m=(2*j-k-g+h)%7
    n=(g-h+m+114)//31
    p=(g-h+m+114)%31
    jour=p+1
    mois=n
    return affichedate([jour,mois,an])

###############################################################################
# Trouve la date du jeudi de l'ascension (paques + 39j)
#  (voir ici: http://fr.wikipedia.org/wiki/Calendrier_chr%C3%A9tien)
#
def ascension(an):
    """ascension(an): trouve la date du jeudi de l'ascension (paques + 39j)"""
    return adddate(paques(an),39)

###############################################################################
# Trouve la date du dimanche de la Pentecôte (Pâques + 49j)
#  (voir ici: http://fr.wikipedia.org/wiki/Calendrier_chr%C3%A9tien)
#
def pentecote(an):
    """pentecote(an): trouve la date du dimanche de la Pentecôte (paques + 49j)"""
    return adddate(paques(an),49)

###############################################################################
# Trouve la date du mardi gras (Pâques - 47j)
#  (voir ici: http://fr.wikipedia.org/wiki/Calendrier_chr%C3%A9tien)
#
def mardigras(an):
    """mardigras(an): trouve la date du mardi gras (Pâques - 47j)"""
    return adddate(paques(an),-47)

###############################################################################
# Trouve la date du dimanche des Rameaux (Pâques - 7j)
#  (voir ici: http://fr.wikipedia.org/wiki/Calendrier_chr%C3%A9tien)
#
def rameaux(an):
    """rameaux(an): trouve la date du dimanche des Rameaux (Pâques - 7j)"""
    return adddate(paques(an),-7)

###############################################################################
# Renvoie la liste des jours fériés (France) de l'année donnée en 1er paramètre
#   si pas de 2ème paramètre (optionnel), les samedis et dimanches sont exclus
#   en ajoutant 1 comme 2ème paramètre, on a tout, samedis et dimanches compris
#   en ajoutant 2 comme 2ème paramètre, on a, en plus, les 2 jours fériés d'Alsace-Moselle
#   à adapter selon votre situation, region, pays, besoins et évolution de la législation
#
def feries(an,sd=0):
    """feries(an,sem=0): => liste des jours fériés (France) de l'année, hors samedi et dimanche. Si sd=1, on les a tous; si sd=2, on a en plus les 2 jours fériés supplémentaires d'Alsace-Moselle"""
    global prog
    # Vérification des données
    if (not prog.estentier(an)) or (an<=0):
        raise ValueError ("erreur feries(an,sd=0): an doit être un nombre entier positif")
    if (not prog.estentier(sd)) or (sd<0) or (sd>2):
        raise ValueError ("erreur feries(an,sd=0): sd doit être un nombre entier = 0, 1 ou 2")
    # Calcul
    F=[]

    X="01/01/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Jour de l'an = " + joursem(X) + " " + X
        F.append(D)

    # Spécifique Alsace-Moselle: le vendredi saint est férié
    if sd==2:
        X=adddate(paques(an),-2)
        D="Vendredi Saint = " + joursem(X) + " " + X
        F.append(D)

    X=paques(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Jour de Pâques = " + joursem(X) + " " + X
        F.append(D)

    X=adddate(paques(an),1)
    D="lundi de Pâques= " + joursem(X) + " " + X
    F.append(D)

    X="01/05/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Fête du Travail = " + joursem(X) + " " + X
        F.append(D)

    X="08/05/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Fête de la Victoire 1945 = " + joursem(X) + " " + X
        F.append(D)

    X=ascension(an)
    D="Ascension = "  + joursem(X) + " " + X
    F.append(D)

    X=pentecote(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Pentecôte = " + joursem(X) + " " + X
        F.append(D)

    X=adddate(pentecote(an),1)
    D="Lundi de Pentecôte = " + joursem(X) + " " + X
    F.append(D)

    X="14/07/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Fête Nationale = " + joursem(X) + " " + X
        F.append(D)

    X="15/08/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Assomption = " + joursem(X) + " " + X
        F.append(D)

    X="01/11/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Toussaint = " + joursem(X) + " " + X
        F.append(D)

    X="11/11/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Armistice 1918 = " + joursem(X) + " " + X
        F.append(D)

    X="25/12/"+str(an)
    if (numjoursem(X)<6) or (sd<>0):
        D="Noël = " + joursem(X) + " " + X
        F.append(D)

    # Spécifique Alsace-Moselle: la St-Etienne, lendemain de Noël, est fériée
    if sd==2:
        X="26/12/"+str(an)
        D="St Etienne = " + joursem(X) + " " + X
        F.append(D)

    return F


